use Sys::Hostname;
use Cwd qw(abs_path);
use File::Basename;
use AnyEvent::Util;
use JSON;
use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
package FastX;
use strict;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(fatal, result, cleanup);

our $api_dir;
BEGIN
{
	$api_dir = Cwd::abs_path(File::Basename::dirname(__FILE__));
}

sub fatal
{
	my %args = @_ unless $_[0] + 0;
	unless (%args) {
		# Old calling format: fatal code, message
		my $code = shift || -32603;
		my $message = shift || "Unknown Error";
		my $extra = shift;
		%args = (code => $code, message => $message);
		%args = (%args, %$extra) if ref $extra;
	}
	if ($args{name} + 0) {
		# Translate a number into an error
		my @name = grep Errno->$_() == $args{name}, keys %!;
		$args{name} = $name[0] if @name;
	}
	print JSON::encode_json( { error => \%args } ), "\n";
	exit 1;
}

sub config_dir
{
	return $ENV{FX_CONFIG_PATH} if $ENV{FX_CONFIG_PATH};
	return Cwd::abs_path(File::Basename::dirname(__FILE__) . '/../config');
}

sub fastx_config
{
	my $config_dir = config_dir();
	my $config_file = $config_dir . '/FastX.conf';
	my $fastx_config = { };
	$fastx_config->{FastXPath} = "$ENV{HOME}/.fastx_server" if $ENV{HOME};
	my $host = ::hostname;
	$fastx_config->{HOST} = $host if $host;
	$fastx_config->{SessionPath} = "$ENV{HOME}/.fastx_server/$host/sessions" if $ENV{HOME} && $host;
	my $user = $ENV{LOGNAME} || ::getpwuid($<);
	$fastx_config->{USER} = $user if $user;
	open my $fh, $config_file or return $fastx_config;
	while (<$fh>) {
		chomp;
		# Comments
		s/^\#.*$//;
		s/([^@])\#[^@].*$/\1/;
		# White space
		s/^\s*//;
		s/\s*$//;
		next if /^$/;
		my ($key, $value) = split /=/, $_, 2;
		$key =~ s/\s*$//;
		$value =~ s/^\s*//;
		$value =~ s/\s*$//;
		next unless $key;
		delete $fastx_config->{key}, next unless $value;
		# Parse variables
		my @vars = m/(\@[^@]*\@)/g;
		my %vars;
		foreach my $var (@vars) {
			my ($v) = $var =~ m/^\@(.*)\@$/;
			$vars{$var} = '@', next unless $v;
			$vars{$var} = '#', next if $v eq '#';
			$vars{$var} = $fastx_config->{$v}, next if $fastx_config->{$v};
			$vars{$var} = $ENV{$v}, next if $ENV{$v};
			$vars{$var} = '';
		}
		foreach my $var (keys %vars) {
			$value =~ s/$var/$vars{$var}/g;
		}
		$fastx_config->{$key} = $value;
	}
	return $fastx_config;
}

sub fastx_dir
{
	my $config = fastx_config();
	return $config->{FastXPath};
}

sub session_dir
{
	my $config = fastx_config();
	return $config->{SessionPath};
}

sub result
{
	my $data = shift || [];
	print STDOUT JSON::encode_json ( { result => $data } ), "\n";
	exit;
}

sub link_ssh
{
	my $sshd_sock = shift;
	my $ssh_auth_sock = shift;
	my $dir = $ssh_auth_sock;
	$dir =~ s|/[^/]*$||;
	mkdir $dir or return unless -d $dir;
	chmod 0700, $dir;
	unlink $ssh_auth_sock;
	symlink $sshd_sock, $ssh_auth_sock;
}

sub command_allowed
{
	my $API = 'command';
	my $command = shift;
	my $profile = shift;
	if (ref($command) eq "HASH") {
		$API = $command->{API} if $command->{API};
		$profile = $command->{profile};
		$command = $command->{command};
	}
	our $patterns;
	our $suggestions;
	unless ($suggestions) {
		$suggestions = { };
		my $confdir = config_dir();
		my ($fh, $list);
		if (open $fh, '<', "$confdir/suggestions.json" and
			$list = eval { JSON::decode_json (join ' ', <$fh>) } and
			ref $list eq 'ARRAY') {
			foreach my $suggestion (@$list) {
				my $key = $suggestion->{profile} or '';
				push @{$suggestions->{$key}}, $suggestion->{command};
			}
		}
	}
	unless ($patterns) {
		$patterns = { };
		my $confdir = config_dir();
		my $fh;
		if (open $fh, '<', "$confdir/$API.json" and
			my $conf = eval { JSON::decode_json (join ' ', <$fh>) } ) {
			$patterns->{allow} = $conf->{allow} if exists $conf->{allow};
			$patterns->{deny} = $conf->{deny} if exists $conf->{deny};
		} else {
		if (open $fh, '<', "$confdir/$API.allow") {
			$patterns->{allow} = [ ];
			while (my $pat = <$fh>) {
				chomp $pat;
				next unless $pat;
				next if $pat =~ /^\#/;
				next if $pat =~ /^\</;	# Discard "<SUGGESTIONS>" and misspellings.
				$pat = "^$pat" unless $pat =~ /^\^/;
				push @{$patterns->{allow}}, $pat;
			}
		}
		if (open $fh, '<', "$confdir/$API.deny") {
			$patterns->{deny} = [ ];
			while (my $pat = <$fh>) {
				chomp $pat;
				next unless $pat;
				next if $pat =~ /^\#/;
				$pat = "^$pat" unless $pat =~ /^\^/;
				push @{$patterns->{deny}}, $pat;
			}
		}
		}
	}
	return 1 if grep $command eq $_, @{$suggestions->{$profile}};

	my $allow;
	if ($patterns->{allow}) {
		my ($pat, $pattern);
		foreach $pat (@{$patterns->{allow}}) {
			next if ref $pat and ref $pat ne 'HASH';
			if ($profile) {
				next unless ref $pat and $profile eq $pat->{profile};
				$pattern = $pat->{regex};
			} else {
				next if ref $pat and $pat->{profile};
				$pattern = ref $pat ? $pat->{regex} : $pat;
			}
			next unless $pattern;
			last if $allow = eval { $command =~ /$pattern/ };
		}
	} elsif (!$profile) {
		$allow = 1;
	}
	if ($patterns->{deny}) {
		my $pat;
		foreach $pat (@{$patterns->{deny}}) {
			undef $allow if eval { $command =~ /$pat/ };
		}
	}
	return $allow;
}

sub bookmark
{
	my $fastx_dir = FastX::fastx_dir;
	my $config_dir = FastX::config_dir;

	my @lists = (
		{ name => 'bookmarks', file => "$fastx_dir/bookmarks", },
		{ name => 'suggestions', file => "$config_dir/suggestions.json", },
		{ name => 'history', file => "$fastx_dir/history", });
	my $list;
	my $result = { };
	for $list (@lists) {
		my $fp;
		my $bookmarks;
		next unless open $fp, '<', $list->{file};
		next unless $bookmarks = eval { JSON::decode_json(join '', <$fp>) };
		$result->{$list->{name}} = $bookmarks;
	}
	return { result => $result };
}

sub session_api
{
	my %args = @_;
	my $request = $args{request};
	my $cb = $args{cb};
	my $action = $request->{action};
	my $filter = $args{filter};
	my @args;
	my $result;
	if ($action) {
		# Backwards compatibility
		$result = { error => { name => 'action_unknown',
			code => -32601,
			message => 'Invalid action' } } if $action =~ m|/|;
		@args = @{$request->{args}} if ref $request->{args} eq 'ARRAY';
		@args = ($request->{args}) if $request->{args} && ! ref $request->{args};
	} else {
		$action = $request->{method};
		$action =~ s/^FastX\.//;
		$action =~ s/^session-//;
		$result = { error => { name => 'api_argument_missing',
			code => -32600, message => 'Missing method' } } unless $action;
		$result = { error => { name => 'method_unknown',
			code => -32601,
			message => 'Invalid action' } } if $action =~ m|/|;
		if ($filter) {
			return &{$filter->{$action}}($request, $cb) if ref $filter->{$action} eq 'CODE';
			$result = { error => { name => 'method_denied',
			message => "Method $action is not allowed." } } unless $filter->{$action};
		}
		if ($action eq 'bookmark' && (!$request->{params} || !$request->{params}->{op} || $request->{params}->{op} eq 'list')) {
			my $result = bookmark($request);
			$result->{msgId} = $request->{msgId} if exists $request->{msgId};
			$result->{id} = $request->{id} if exists $request->{id};
			$result->{jsonrpc} = $request->{jsonrpc} if exists $request->{jsonrpc};
			return &$cb($result);
		}
		if ($action =~ m/^(bookmark)-(.*)/) {
			$action = $1;
			@args = ("--op=$2");
		}
		while ((my $key, my $value) = each %{$request->{params}}) {
			if (ref $value eq 'HASH') {
				# Ignore key name (which will probably be 'params').
				while ((my $key, my $v) = each %$value) {
					$key .= "=$v";
					push @args, $key;
				}
			} else {
				$key = "--$key" unless $action eq 'params' && $key ne 'id'
					or $action eq 'bookmark' && $key ne 'op';
				$key .= "=$value";
				push @args, $key;
			}
		}
	}
	if ($result) {
		$result->{msgId} = $request->{msgId} if exists $request->{msgId};
		$result->{id} = $request->{id} if exists $request->{id};
		$result->{jsonrpc} = $request->{jsonrpc} if exists $request->{jsonrpc};
		return &$cb($result);
	}
	my $cv = AnyEvent::Util::run_cmd [ "$api_dir/$action", @args ],
		'>' => \my $output, '2>' => \my $error, '<' => '/dev/null';
	$cv->cb(sub {
		my $status = shift->recv;
		my $result = eval { JSON::decode_json $output };
		if (ref $result ne 'HASH') {
			$result->{error} = { name => 'action_failed', code => -32001,
				message => "exit status $status, error $error, json $@" };
		}
		$result->{msgId} = $request->{msgId} if exists $request->{msgId};
		$result->{id} = $request->{id} if exists $request->{id};
		$result->{jsonrpc} = $request->{jsonrpc} if exists $request->{jsonrpc};
		&$cb($result);
	});
	return undef;
}

package CandC;
use Socket;
use AnyEvent::Socket;
sub new
{
	my $class = shift;
	my $self = { active => [ ], id => 0 };
	bless $self, $class;
	return $self;
}

sub DESTROY
{
	my $self = shift;

	$self->disconnect();
}

sub connect
{
	my $self = shift;
	my $display = shift;
	my $cb = shift;
	my $cv = AE::cv unless $cb;
	$self->{display} = $display if defined $display;
	return &$cb(), undef unless defined $self->{display};
	$self->{file} = "/tmp/Xrdp_control_$display.sock";
	my $w = tcp_connect 'unix/', $self->{file}, sub {
		$self->{sock} = shift;
		$self->{is_connected} = $self->{sock};
		&$cb($self->{sock}, @_) if $cb;
		$cv->send($self->{sock}) if $cv;
		return unless $self->{sock};
		my $len;
		my $buf;
		$self->{io} = AE::io($self->{sock}, 0, sub {
			my $n;
			unless (defined $len) {
				while ($n = sysread($self->{sock}, $buf, 1, length $buf)) {
					if (substr($buf, -1) eq ':') {
						chop $buf;
						$len = $buf + 1; # Add 1 for comma.
						$buf = '';
						last;
					}
				}
			}
			if (defined $len) {
				while ($len && ($n = sysread($self->{sock}, $buf, $len, length $buf))) {
					$len -= $n;
				}
				if ($len == 0) {
					undef $self->{timeout};
					chop $buf;	# Remove comma
					$buf = eval { JSON::decode_json $buf };
					my $rpc = shift @{$self->{active}};
					&{$rpc->{cb}}($buf) if $rpc;
					$self->{timeout} = AE::timer(10, 0, sub { $self->disconnect({ error => { code => -32010, message => 'CandC timeout' } }); }) if $self->{active};
					undef $buf;
					undef $len;
					return;
				}
			}
			if (!$n) {
				return if !defined $n && $!{EAGAIN};
				unlink($self->{file}), undef($self->{file});
				$self->disconnect();
			}
		});
	};
	return $cv->recv if $cv;
	return $w;
}

sub disconnect
{
	my $self = shift;
	my $result = shift;

	undef $self->{sock};
	undef $self->{is_connected};
	undef $self->{io};
	undef $self->{timeout};
	foreach my $rpc (@{$self->{active}}) {
		&{$rpc->{cb}}($result);
	}
	$self->{active} = [];
}

sub set_permission
{
	my $self = shift;
	my $perm = shift || 0700;
	chmod $perm, $self->{file};
}

sub _process_status
{
	my $self = shift;
	my $status = shift;
	my $key;
	$status = $status->{result} or return undef;
	foreach $key (keys %$status) {
		my $lkey = lcfirst $key;
		next if $key eq $lkey;
		$status->{$lkey} = delete $status->{$key};
	}
	$status->{clients} = 0 + $status->{clients} if defined $status->{clients};
	if (defined $status->{screenWidth} && defined $status->{screenHeight}) {
		$status->{geometry} = "$status->{screenWidth}x$status->{screenHeight}";
		delete $status->{screenWidth};
		delete $status->{screenHeight};
	}
	return $status;
}

sub _rpc_ae
{
	my $self = shift;
	my $method = shift;
	my $params = shift;
	my $cb = shift;
	$cb = $params, undef $params if ref $params eq 'CODE';
	my $call = { method => $method, msgId => ++$self->{id} };
	$call->{params} = $params if $params;
	my $data = JSON::encode_json $call;
	my $net_string = length $data;
	$net_string .= ":$data,";

	push @{$self->{active}}, { cb => $cb, id => $self->{id} };
	$self->disconnect(), return undef unless $self->{sock};
	unless (syswrite $self->{sock}, $net_string, length $net_string) {
		unlink($self->{file}), undef($self->{file});
		$self->disconnect();
		return undef;
	}
	$self->{timeout} = AE::timer(10, 0, sub { $self->disconnect({ error => { code => -32010, message => 'CandC timeout' } }); });
	return $self->{id};
}

sub _rpc
{
	my $self = shift;
	my $method = shift;
	my $cb = pop;
	if (defined $cb && ref $cb ne 'CODE') {
		push @_, $cb;
		undef $cb;
	}
	my $cv = AE::cv unless $cb;
	$self->_rpc_ae($method, sub {
		&$cb(@_) if $cb;
		$cv->send(@_) if $cv;
	});
	return $cv->recv if $cv;
}

sub version
{
	my $self = shift;
	my $cb = shift;
	my $cv = AE::cv unless $cb;
	$self->_rpc_ae('getVersion', sub {
		my $result = shift->{result};
		&$cb($result) if $cb;
		$cv->send($result) if $cv;
	});
	return $cv->recv if $cv;
}

sub get_status
{
	my $self = shift;
	my $cb = shift;
	my $cv = AE::cv unless $cb;
	$self->_rpc_ae('getStatus', sub {
		my $data = shift;
		&$cb($self->_process_status($data)) if $cb;
		$cv->send($self->_process_status($data)) if $cv;
	});
	return $cv->recv if $cv;
}

sub get_password
{
	my $self = shift;
	my $params = shift;
	my $cb = shift;
	$cb = $params, undef $params if ref $params eq 'CODE';
	my $cv = AE::cv unless $cb;
	$self->_rpc_ae('getPassword', $params, sub {
		my $result = shift->{result};
		my $password = $result && $result->{Password};
		&$cb($password) if $cb;
		$cv->send($password) if $cv;
	});
	return $cv->recv if $cv;
}

sub checkout_license
{
	my $self = shift;
	unshift @_, $self, 'checkoutLicense';
	&_rpc;
}

sub get_license_status
{
	my $self = shift;
	unshift @_, $self, 'getLicenseStatus';
	&_rpc;
}

sub reset
{
	my $self = shift;
	unshift @_, $self, 'reset';
	&_rpc;
}

sub reset_sharing
{
	my $self = shift;
	unshift @_, $self, 'resetSharing';
	&_rpc;
}

sub shutdown
{
	my $self = shift;
	unshift @_, $self, 'shutdown';
	&_rpc;
}

# Check out a license and return success or failure.
# If return object has an "error" element, this typically means
# there was a communication failure (not a license failure).
# A license failure is indicated by a "result" element that doesn't
# have a "true" value in "LicenseIsGood". In that case, the
# license error (string from RLM) will be in the "LastError" element of "result".
# Since the string from RLM is typically not useful, we parse it
# here and return something useful.
sub get_license
{
	my $self = shift;
	my @backoff = ( 0.2, 0.5, 1, 2, 5, 10, 20 );
	my $time;

	$self->checkout_license or return { error => { message => "CandC failure.", name => 'candc_error' } };
	for $time (@backoff) {
		my $license = $self->get_license_status or return { error => { message => 'CandC failure.', name => 'candc_error' } };
		return $license if $license->{error};
		return { error => { message => 'Unexpected license result', name => 'license_error' } } unless $license->{result};
		return $self->translate_license_error($license) unless $license->{result}->{CheckoutIsRunning};
		select undef, undef, undef, $time;
	}
	return { error => { message => 'License timeout.', name => 'license_timeout'} };
}

sub translate_license_error
{
	my $self = shift;
	my $license = shift;

	return $license unless $license->{result};
	my $LastError = $license->{result}->{LastError};
	return $license unless $LastError; # Success!
	return $license if $license->{result}->{LicenseIsGood};
	# Error may be multi-line, but we only care about the last one
	my @list = split /\n/, $LastError;
	$LastError = pop @list;
	# RLM string errors have the code in parenthesis.
	my $code;
	my $name = 'license_error';
	($LastError, $code) = ($LastError =~ /(.*)\s+\(([0-9\-]+)\)$/);
	$LastError = "A FastX license is not installed.", $name = 'license_none' if $code == -1;
	$LastError = "A FastX license is required, but not found.", $name = 'license_none' if $code == -18;
	$LastError = "License server is not running", $name = 'license_server_down' if $code == -111;
	$LastError = "Your license is too old for this version of FastX.", $name = 'license_old' if $code == -6;
	$LastError = "Bad hostname in license file.", $name = 'license_bad' if $code == -131;
	$LastError = "License error: $LastError" if $LastError and $LastError !~ /license/i;
	$LastError = "Unknown license error" unless $LastError;
	delete $license->{result};
	$license->{error} = { name => $name, message => $LastError };
	$license->{error}->{LicenseErrorCode} = $code if $code;

	return $license;
}
