#!/bin/bash
INSTALLDIR=`dirname $0`
INSTALLDIR=`cd "$INSTALLDIR"; pwd`
VAR="$INSTALLDIR/var"
LICENSEDIR="$VAR/license"
set -o pipefail
QUIET=
ARGS=$*
while [[ $# > 0 ]]; do
key="$1"
case "$key" in
        -q|--quiet)
                QUIET="$key"
                ;;
esac
shift
done

umask +rx
yesno() {
answer=
while [[ ! $answer =~ ^[nNyY] ]]; do
	read -p "$1" answer || exit 1
	[ -z "$answer" ] && answer=$2
done
[[ $answer =~ ^[yY] ]] || answer=
}

owner=
who_owns() {
	owner=`ls -ld "$1" | awk 'NR==1 {print $3}'`
}

if [ `whoami` == 'root' ]; then
	ROOT=1
else
	ROOT=
	if [ ! -w "$INSTALLDIR" ]; then
		echo "This install script must be run as root."
		exit 1
	fi
	echo "Warning: This script has limited functionality when not running as root."
	echo "Non-root installation instructions can be found in at"
	echo "http://www.starnet.com/xwin32kb/fastx-nonroot-installation/."
fi

# The following lines are also in install/suggestions, which should have been run first.
# In case the RPM was installed with --noscripts, we do it here again.
if [ -d "$INSTALLDIR/config" -a ! -L "$INSTALLDIR/config" -a -d "$VAR/config" ]; then
	# Both old (config) and new (var/config) directories exist. This shouldn't happen.
	echo "Warning: Both $INSTALLDIR/config and $VAR/config directories exist."
elif [ -d "$VAR/config" ]; then
	# new (config) exists. The old is probably already a symlink, but let's make sure.
	[ -L "$INSTALLDIR/config" ] || ln -s var/config "$INSTALLDIR/config"
elif [ -d "$INSTALLDIR/config" ]; then
	# Upgrade from old (2.3) version. Move config to new location.
	who_owns "$INSTALLDIR/config"
	mkdir "$VAR" 2>/dev/null && [ -n "$owner" ] && chown "$owner" "$VAR"
	mv "$INSTALLDIR/config" "$VAR/config" && ln -s var/config "$INSTALLDIR/config"
else
	# New installation. Make new var/config and symlink for api scripts. We don't know the owner yet.
	mkdir "$VAR" && mkdir "$VAR/config" && ln -s var/config "$INSTALLDIR/config"
fi

# Upgrade: Move license files to new $LICENSEDIR
for f in "$INSTALLDIR"/rlm/*.lic; do
	if [ -e "$f" ]; then
		OLD_LICENSE=1
	else
		OLD_LICENSE=
	fi
	break
done
for f in "$LICENSEDIR"/*.lic; do
	if [ -e "$f" ]; then
		NEW_LICENSE=1
	else
		NEW_LICENSE=
	fi
	break
done
if [ -n "$OLD_LICENSE" -a -z "$NEW_LICENSE" ]; then
	# We have license in the old location, but not the new location.
	# This installation must be an upgrade.
	# We'll create the new var directory, making it owned by the same user as the old directory,
	# which should be 'fastx'. If we're wrong, the perl install scripts will fix it.
	who_owns "$VAR" || who_owns "$INSTALLDIR/rlm" || who_owns "$INSTALLDIR/config"
	[ -n "$owner" ] && chown "$owner" "$VAR"
	mkdir "$LICENSEDIR" && [ -n "$owner" ] && chown "$owner" "$LICENSEDIR"
	mv "$INSTALLDIR"/rlm/*.lic "$LICENSEDIR/" && (cd "$INSTALLDIR/rlm"; ln -s ../var/license/*.lic .)
fi

if [ -d "$INSTALLDIR/certs" -a ! -d "$VAR/certs" ]; then
	[ -n "$owner" ] || who_owns "$INSTALLDIR/certs"
	mv "$INSTALLDIR/certs" "$VAR/certs" && ln -s var/certs "$INSTALLDIR/certs"
fi

# Does a license file already exist?
if [ -n "$OLD_LICENSE" -o -n "$NEW_LICENSE" ]; then
	"$INSTALLDIR"/install/activate --reinstall
	if [ -n "$QUIET" ]; then
		DO_LICENSE=
	else
		echo "There is already a license file in $LICENSEDIR."
		yesno "Would you like to activate a new license? [y/N] " n
		DO_LICENSE=$answer
	fi
elif [ -n "$QUIET" ]; then
	# If we have to be quiet, then we can't prompt to install a new license.
	DO_LICENSE=
else
	DO_LICENSE=1
fi

if [ "$DO_LICENSE" ]; then
	yesno "Do you already have a license server on your network? [y/N] " n
	if [ "$answer" ]; then
		answer=
		while [ -z "$answer" ]; do
			read -p "Enter the name (or IP address) of your license server: " answer extra || exit 1
		done
		[ -d "$LICENSEDIR" ] || mkdir -p "$LICENSEDIR" || exit
		echo "HOST $answer 00000000 5053" > "$LICENSEDIR/$answer.lic" || exit
		echo "License file $LICENSEDIR/$answer.lic has been created."
	else
		yesno "Would you like to activate a license now? [Y/n] " y
		if [ "$answer" ] && ! "$INSTALLDIR"/install/activate; then
			yesno "Continue? [Y/n] " y
			[ "$answer" ] || exit 1
		fi
	fi
fi
# License done - proceed to web
if [ -n "$QUIET" ]; then
	"$INSTALLDIR"/install/web "$QUIET"
	"$INSTALLDIR"/install/screenshot "$QUIET"
else
	yesno "Install/update the FastX web server? [Y/n] " y
	[ "$answer" ] && "$INSTALLDIR"/install/web
	[ "$answer" ] || "$INSTALLDIR"/install/web -r
# Check screenshot support
	if ! "$INSTALLDIR"/install/screenshot --query; then
		yesno "Install screenshot support? [Y/n] " y
		if [ "$answer" ]; then
			"$INSTALLDIR"/install/screenshot
		else
			"$INSTALLDIR"/install/screenshot --no
		fi
	fi
fi
# The suggestions file should have already been created on install,
# but on Debian systems, the 'alien' utility does not run the installation
# script by default, so let's run it here just to make sure.
"$INSTALLDIR"/install/suggestions -q
