'use strict';

var DataStream  = require('DataStream');
var tpkt        = require('xrdp/core/tpkt');
var x224        = require('xrdp/core/x224');
var per         = require('xrdp/core/per');
var ber         = require('xrdp/core/ber');


function writeDomainMCSPDUChoice(num){
    if(num >= 0 && num <= 42){
        return num << 2;
    }
}
function readDomainParameters(dv,offset){//client ignores all data
    var o = ber.readInteger(dv,offset);
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //maxUserIds
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //maxTokenIds
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //numPriorities
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //minThroughput
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //maxHeight
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //maxMCSPDUsize
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset); //protocolVersion
    if(!o){return undefined;}
    return offset + o.consumed;
}

function readMcsCrsp(dv){
    var o,offset = 0;
    o = ber.readAndCheckApplicationTag(dv,offset,102);
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readEnumerated(dv,offset);
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readInteger(dv,offset);
    if(!o){return undefined;}
    offset += o.consumed;
    o = ber.readAndCheckSequenceTag(dv,offset);
    if(!o){return undefined;}
    offset += o.consumed;
    return readDomainParameters(dv,offset);
}

function readGccConnectionData(dv,offset){
    var o; //ds.readUint8();
    o = ber.readLength(dv,++offset);
    if(!o){return undefined;}
    offset += o.consumed;
    if(per.readAndCheckObjectIdentifier(dv,++offset,[0,0,20,124,0,1])!== 1){
        throw new Error( "Invalid object Identifier" );
    }
    return offset + 6;
}

function readGccCCrsp(dv,offset){
    if(dv.getUint8(offset++) !== 0x14){
        throw new Error("Invalid choice with User Data Present");
    }
    var nodeId = dv.getUint16(offset,true);
    if(nodeId < 1001 && nodeId > 65535){
        throw new Error("Invalid NodeId");
    }
    offset +=2;
    var o= per.readInteger(dv,offset);//tag
    if(!o){return undefined;}
    offset += o.consumed +  4;
    if( dv.getUint8(offset) !== 77   ||
        dv.getUint8(offset+1) !== 99 ||
        dv.getUint8(offset+2) !== 68 ||
        dv.getUint8(offset+3) !== 110 ){
        throw new Error('Invalid H.221 Server Identifier');
            //server-to-client H.221 key "McDn" -- [77,99,68,110]
    }
    return offset + 4;
}

function readServerCoreData(dv,offset,store){
    if(dv.getUint16(offset,true) !== 0x0C01){
        throw new Error("Invalid Server Core Data Header");
    }
    var len = dv.getUint16(offset +2,true);
    if(len > 0){
        store.receivedServerData.version = dv.getUint32(offset +4 ,true); //version 0x00080004
    }
    if(len > 4){
        store.receivedServerData.clientRequestedProtocol = dv.getUint32(offset + 8 ,true); //clientRequestedProtocol
    }
    if(len > 8){
        store.receivedServerData.earlyCapabilitiesFlags = dv.getUint32(offset + 12,true); //earlyCapabilitiesFlags -- optional?
    }
    return offset +len;
}

function readServerNetworkData(dv,offset,store){
    if(dv.getUint16(offset,true) !== 0x0C03){
        throw new Error("Invalid Server Network Data header");
    }
    var len = dv.getUint16(offset +2 ,true);
    if(len % 4 !== 0){
        throw new Error('Invalid Server Network Data Length');
    }
    store.ioChannelId = dv.getUint16(offset + 4,true);
    var channelCount = dv.getUint16(offset + 6 ,true);
    for(var i=0;i < channelCount; i++){
        store.staticVirtualChannelIds.push(dv.getUint16(offset +8 +i*2,true));
    }
    return offset + len;
}


function decodeServerCertificateData(dv,offset,certLen){
    var o = {}
    var ds = new DataStream(dv.buffer,dv.byteOffset + offset);
    var byte = ds.readUint32();
    o.certChainVersion =  ( byte & 0x7FFFFFFF) ;
    o.certTemporarilyIssued = (byte & 0x80000000) >>> 31;
    o.data = {};
    switch(o.certChainVersion){
        case 0x00000001:
            if(ds.readUint32()!== 0x00000001){
                throw new Error('Bad Signature Algorithm Identifier');
            }
            if(ds.readUint32()!== 0x00000001){
                throw new Error('Bad Key Algorithm Identifier');
            }
            if(ds.readUint16() !== 0x0006){
                throw new Error('Bad Public Key Blob Type');
            }
            o.data.publicKey = ds.readUint8Array(ds.readUint16());
            if(ds.readUint16()!== 0x0008){
                throw new Error('Bad Signature Blob Type');
            }
            o.data.signature = ds.readUint8Array(ds.readUint16());
            return o;
            break;
        case 0x00000002:
            break;
        default:
            throw new Error('Invalid Cert Chain Version');
    }
}

function readServerSecurityData(dv,offset,store){
    if(dv.getUint16(offset,true) !== 0x0C02){
        throw new Error("Invalid Server Security Data Header");
    }
    offset +=2;
    var secDataLen = dv.getUint16(offset,true);
    offset +=2;
    store.receivedServerData.encryptionMethod = dv.getUint32(offset,true);//encryptionMethod
    if([0x00000000,0x00000001,0x00000002,0x00000008,0x00000010].indexOf(store.receivedServerData.encryptionMethod) === -1){
        throw new Error('Invalid Encryption Method');
    }
    offset +=4;
    store.receivedServerData.encryptionLevel = dv.getUint32(offset,true);//encryptionLevel
    offset +=4;
    if(store.receivedServerData.encryptionMethod || store.receivedServerData.encryptionLevel){
        throw new Error("Encryption Level not set to 0");
        var randLen = dv.getUint32(offset,true);
        if(0x00000020 !== randLen){
            throw new Error('Invalid serverRandomLen');
        }
        offset+=4;
        var certLen = dv.getUint32(offset,true);
        offset+=4;
        store.receivedServerData.serverRandom = new DataView(dv.buffer,dv.byteOffset+offset,randLen);
        offset+=randLen;
        store.receivedServerData.serverCertificate = decodeServerCertificateData(dv,offset,certLen);
        offset+=certLen;
    }
    return offset;
}

function readServerMessageChannelData(dv,offset,store){
    if(dv.getUint16(offset,true) !== 0x0C04){
        throw new Error("Invalid Server Message Channel Data Header");
    }
    offset +=2;
    var len = dv.getUint16(offset,true);
    offset += 2;
    store.messageChannelId = dv.getUint16(offset,true);
    offset += 2;
    return offset;
}

function readServerMultiTransportChannelData(dv,offset){
    if(dv.getUint16(offset,true) !== 0x0C08){
        throw new Error("Invalid Server Multitransport Channel Data Header");
    }
    offset +=2;
    offset += dv.getUint16(offset,true);
    return offset;
}

function decodeConferenceCreateResponse(dv,store,messageChannelData,multiTransportData){
    var o;
    var offset = readMcsCrsp(dv);
    if(!offset){throw new Error('readMcsCrsp failed');}
    offset = readGccConnectionData(dv,offset);
    if(!offset){throw new Error('readGccConnectionData failed');}
    o = per.readLength(dv,offset);
    if(!o){throw new Error('readLength failed');}
    offset += o.consumed;
    offset = readGccCCrsp(dv,offset);
    if(!offset){throw new Error('readGccCCrsp failed');}
    o = per.readLength(dv,offset);
    if(!o){throw new Error('readLength failed');}
    offset += o.consumed;
    offset = readServerCoreData(dv,offset,store);
    if(!offset){throw new Error('readServerCoreData failed');}
    offset = readServerNetworkData(dv,offset,store);
    if(!offset){throw new Error('readServerNetworkData failed');}
    offset = readServerSecurityData(dv,offset,store);
    if(!offset){throw new Error('readServerSecurityData failed');}
    offset = ((typeof messageChannelData !== 'undefined')?readServerMessageChannelData(dv,offset,store):offset);
    if(!offset){throw new Error('readServerMessageChannelData failed');}
    return ((typeof multiTransportChannel !== 'undefined')?readServerMultiTransportChannelData(dv,offset):offset);
}

function erectDomainRequest(subHeight,subInterval){
    var ds = new DataStream();
    ds.writeUint8(writeDomainMCSPDUChoice(1));
    per.writeInteger(ds,subHeight);
    per.writeInteger(ds,subInterval);
    return tpkt.writePacket(x224.writeDataPDU(ds.buffer));
}

function attachUserRequest(){
    var ds = new DataStream();
    ds.writeUint8(writeDomainMCSPDUChoice(10));
    return tpkt.writePacket(x224.writeDataPDU(ds.buffer));
}

function processBasicSettingsExchange(socket,Client,dv){
    var data = tpkt.parsePacket(dv);
    if(!data.payload){return 0;}
    data = x224.parseConnectPDU(data.payload);
    if(!data.data){return 0;}
    decodeConferenceCreateResponse(data.data,Client.store,Client.settings.messageChannelData,Client.settings.multiTransportData);
    socket.send(erectDomainRequest(0,0));
    socket.send(attachUserRequest());
    return dv.getUint16(2,false);
}
module.exports = {
    process:processBasicSettingsExchange
}
