'us strict';
/*
 * BER
 */

var DataStream = require('DataStream');
    
module.exports = {
	BER_CLASS_MASK: 0xC0,
	BER_CLASS_UNIV: 0x00,
	BER_CLASS_APPL: 0x40,
	BER_CLASS_CTXT: 0x80,
	BER_CLASS_PRIV: 0xC0,

	BER_PC_MASK: 0x20,
	BER_PRIMITIVE: 0x00,
	BER_CONSTRUCT: 0x20,

	BER_TAG_MASK: 0x1F,
	BER_TAG_BOOLEAN: 0x01,
	BER_TAG_INTEGER: 0x02,
	BER_TAG_BIT_STRING: 0x03,
	BER_TAG_OCTET_STRING: 0x04,
	BER_TAG_OBJECT_IDENTIFIER: 0x06,
	BER_TAG_ENUMERATED: 0x0A,
	BER_TAG_SEQUENCE: 0x10,
	BER_TAG_SEQUENCE_OF: 0x10,

	getPC: function(pc) {
		return pc ? this.BER_CONSTRUCT : this.BER_PRIMITIVE;
	},

	readLength: function(dv, offset) {
		var length = dv.getUint8(offset);
		if(!(length & 0x80)){return { length: length, consumed: 1 };}
		switch (length) {
		case 0x81:
			return { length: dv.getUint8(offset + 1), consumed: 2 };
		case 0x82:
			return { length: dv.getUint16(offset + 1,false), consumed: 3 };
		default:
			return undefined;
		}
	},
	writeLength: function(ds, length) {
		if (length > 0xFF) {
			ds.writeUint8(0x82);
			ds.writeUint16(length,DataStream.BIG_ENDIAN);
		}
		else if (length > 0x7F) {
			ds.writeUint8(0x81);
			ds.writeUint8(length);
		}
		else
			ds.writeUint8(length);
	},
	readAndCheckUniversalTag: function(dv, offset, tag, pc) {
		if (dv.getUint8(offset) != (this.BER_CLASS_UNIV | this.getPC(pc) | (this.BER_TAG_MASK & tag)))
			return undefined;

		return { consumed: 1 };
	},
	writeUniversalTag: function(ds, tag, pc) {
		ds.writeUint8(this.BER_CLASS_UNIV | this.getPC(pc) | (this.BER_TAG_MASK & tag));
	},
	readAndCheckApplicationTag: function(dv, offset, tag) {
		if (tag > 30) {
			if (dv.getUint8(offset) != (this.BER_CLASS_APPL | this.BER_CONSTRUCT | this.BER_TAG_MASK))
				return undefined;

			if (dv.getUint8(offset + 1) != tag)
				return undefined;

			var pl = this.readLength(dv,offset + 2);

			if (!pl)
				return undefined;

			pl.consumed += 2;
			return pl;
		}
		else {
			if (dv.getUint8(offset) != (this.BER_CLASS_APPL | this.BER_CONSTRUCT | (this.BER_TAG_MASK & tag)))
				return undefined;

			var pl = this.readLength(dv,offset + 1);

			if (!pl)
				return undefined;

			pl.consumed++;
			return pl;
		}
	},
	writeApplicationTag: function(ds, tag, length) {
		if (tag > 30) {
			ds.writeUint8(this.BER_CLASS_APPL | this.BER_CONSTRUCT | this.BER_TAG_MASK);
			ds.writeUint8(tag);
		}
		else
			ds.writeUint8(this.BER_CLASS_APPL | this.BER_CONSTRUCT | (this.BER_TAG_MASK & tag));

		this.writeLength(ds,length);
	},
	readAndCheckContextualTag: function(dv, offset, tag, pc) {
		if (dv.getUint8(offset) != (this.BER_CLASS_CTXT | this.getPC(pc) | (this.BER_TAG_MASK & tag)))
			return undefined;

		var pl = this.readLength(dv,offset + 1);

		if (!pl)
			return undefined;

		pl.consumed++;
		return pl;
	},
	writeContextualTag: function(ds, tag, pc, length) {
		ds.writeUint8(this.BER_CLASS_CTXT | this.getPC(pc) | (this.BER_TAG_MASK & tag));
		ds.writeLength(length);
	},
	readAndCheckSequenceTag: function(dv, offset) {
		if (dv.getUint8(offset) != (this.BER_CLASS_UNIV | this.BER_CONSTRUCT | this.BER_TAG_SEQUENCE_OF))
			return undefined;

		var pl = this.readLength(dv,offset + 1);

		if (!pl)
			return undefined;

		pl.consumed++;
		return pl;
	},
	writeSequenceTag: function(ds, length) {
		ds.writeUint8(this.BER_CLASS_UNIV | this.BER_CONSTRUCT | this.BER_TAG_SEQUENCE_OF);
		this.writeLength(length);
	},
	readBoolean: function(dv, offset) {
		var consumed = 0;
		var pl = this.readAndCheckUniversalTag(dv,offset + consumed,this.BER_TAG_BOOLEAN,false);

		if (!pl)
			return undefined;

		consumed += pl.consumed;

		pl = this.readLength(dv,offset + consumed);

		if (!pl)
			return undefined;

		consumed += pl.consumed;

		return { value: ds.readUint8(offset + consumed) ? true : false, consumed: consumed + 1 };
	},
	writeBoolean: function(ds, boolean) {
		this.writeUniversalTag(ds,this.BER_TAG_BOOLEAN,false);
		this.writeLength(1);
		ds.writeUint8(boolean ? 0xFF : 0);
	},
	readInteger: function(dv, offset) {
		var consumed = 0;
		var pl = this.readAndCheckUniversalTag(dv,offset + consumed,this.BER_TAG_INTEGER,false);

		if (!pl)
			return undefined;

		consumed += pl.consumed;

		pl = this.readLength(dv,offset + consumed);

		if (!pl)
			return undefined;

		consumed += pl.consumed;

		switch (pl.length) {
		case 1:
			return { value: dv.getUint8(offset + consumed), consumed: consumed + 1 };
		case 2:
			return { value: dv.getUint16(offset + consumed,false), consumed: consumed + 2 };
		case 3:
			return { value: (dv.getUint8(offset + consumed) << 16) | dv.getUint16(offset + consumed + 1,false), consumed: consumed + 3 };
		case 4:
			return { value: dv.getUint32(offset + consumed,false), consumed: consumed + 4 };
		default:
			return undefined;
		}
	},
	writeInteger: function(ds, integer) {
		if (integer < 0x80) {
			this.writeUniversalTag(ds,this.BER_TAG_INTEGER,false);
			this.writeLength(ds,1);
			ds.writeUint8(integer);
		}
		else if (integer < 0x8000) {
			this.writeUniversalTag(ds,this.BER_TAG_INTEGER,false);
			this.writeLength(ds,2);
			ds.writeUint16(integer,DataStream.BIG_ENDIAN);
		}
		else if (integer < 0x800000) {
			this.writeUniversalTag(ds,this.BER_TAG_INTEGER,false);
			this.writeLength(ds,3);
			ds.writeUint8(integer >>> 16);
			ds.writeUint16(integer & 0xFFFF,DataStream.BIG_ENDIAN);
		}
		else if (integer < 0x80000000) {
			this.writeUniversalTag(ds,this.BER_TAG_INTEGER,false);
			this.writeLength(ds,4);
			ds.writeUint32(integer,DataStream.BIG_ENDIAN);
		}
	},
	writeOctetString: function(ds, array) {
		this.writeUniversalTag(ds,this.BER_TAG_OCTET_STRING,false);
		this.writeLength(ds,array.length);
		ds.writeUint8Array(array);
	},
	readEnumerated: function(dv, offset) {
		var consumed = 0;
		var pl = this.readAndCheckUniversalTag(dv,offset + consumed,this.BER_TAG_ENUMERATED,false);

		if (!pl)
			return undefined;

		consumed += pl.consumed;

		pl = this.readLength(dv,offset + consumed);

		if (!pl)
			return undefined;
		if (pl.consumed != 1)
			return undefined;

		consumed += pl.consumed;

		return { value: dv.getUint8(offset + consumed), consumed: consumed + 1 };
	},
	writeEnumerated: function(ds, enumerated) {
		this.writeUniversalTag(ds,this.BER_TAG_ENUMERATED,false);
		this.writeLength(ds,1);
		ds.writeUint8(enumerated);
	}
}

