'use strict';
/*
 * PER
 */
var _           = require('underscore');
var DataStream  = require('DataStream');
    
module.exports = {
	readLength: function(dv, offset) {
		var length = dv.getUint8(offset);
		var consumed = 1;

		if (length & 0x80) {
			length = ((length & ~(0x80)) << 8) | dv.getUint8(offset + 1);
			++consumed;
		}

		return { length: length, consumed: consumed };
	},
	writeLength: function(ds, length) {
		if (length > 0x7F)
			ds.writeUint16(length | 0x8000,DataStream.BIG_ENDIAN);
		else
			ds.writeUint8(length);
	},
	readInteger: function(dv, offset) {
		var pl = this.readLength(dv,offset);

		if (!pl) return undefined;

		switch (pl.length) {
		case 1:
			return { value: dv.getUint8(offset + pl.consumed), consumed: pl.consumed + 1 };
		case 2:
			return { value: dv.getUint16(offset + pl.consumed,false), consumed: pl.consumed + 2 };
		case 4:
			return { value: dv.getUint32(offset + pl.consumed,false), consumed: pl.consumed + 4 };
		default:
			return undefined;
		}
	},
	writeInteger: function(ds, integer) {
		if (integer <= 0xFF) {
			this.writeLength(ds,1);
			ds.writeUint8(integer);
		}
		else if (integer <= 0xFFFF) {
			this.writeLength(ds,2);
			ds.writeUint16(integer,DataStream.BIG_ENDIAN);
		}
		else if (integer <= 0xFFFFFFFF) {
			this.writeLength(ds,4);
			ds.writeUint32(integer,DataStream.BIG_ENDIAN);
		}
	},
	readAndCheckObjectIdentifier: function(dv, offset, oids) {
		var pl = this.readLength(dv,offset);

		if (!pl) return undefined;

		if (pl.length != 5) return undefined;

		var t12 = dv.getUint8(offset + 1);
		var a_oids = [ t12 >>> 4, t12 & 0x0F,
			dv.getUint8(offset + 2),
			dv.getUint8(offset + 3),
			dv.getUint8(offset + 4),
			dv.getUint8(offset + 5) ];
		return _.isEqual(oids,a_oids) ? 1 : undefined;
	},
	writeObjectIdentifier: function(ds, oids) {
		var t12 = (oids[0] << 4) | (oids[1] & 0x0F);

		this.writeLength(ds,5);
		ds.writeUint8(t12);
		ds.writeUint8(oids[2]);
		ds.writeUint8(oids[3]);
		ds.writeUint8(oids[4]);
		ds.writeUint8(oids[5]);
	},
	readAndCheckOctetString: function(dv, offset, array, minimumLength) {
		minimumLength = minimumLength || 0;

		var pl = this.readLength(dv,offset);

		if (!pl) return undefined;

		if (pl.length + minimumLength != array.length) return undefined;

		var ta = new Uint8Array(array.length);

		for (var n = 0; n < array.length; n++) {
			ta[n] = dv.getUint8(offset + pl.consumed + n);
		}

		return _.isEqual(array,ta) ? 1 : undefined;
	},
	writeOctetString: function(ds, array, minimumLength) {
		minimumLength = minimumLength || 0;

		var ml = array.length - minimumLength >= 0
			? array.length - minimumLength
			: minimumLength;

		this.writeLength(ds,ml);
		ds.writeUint8Array(array);
	},
	writeNumericString: function(ds, array, minimumLength) {
		minimumLength = minimumLength || 0;

		var ml = array.length - minimumLength >= 0
			? array.length - minimumLength
			: minimumLength;

		this.writeLength(ds,ml);

		for (var i = 0; i < array.length; i += 2) {
			var c1 = array[i];
			var c2 = i + 1 < array.length ? array[i + 1] : 0x30;

			c1 = (c1 - 0x30) % 10;
			c2 = (c2 - 0x30) % 10;
			ds.writeUint8((c1 << 4) | c2);
		}
	}
}
