'use strict';

    var DataStream      = require('DataStream');
    var Orders          = require('./orders/orders');
    var PointerUpdate   = require('./pointer-update');
    var Surface         = require('./surface/surface');
    var Palette         = require('./palette');
    var Bitmap          = require('./bitmap');
    var CanvasOrder     = require('./orders/canvas-order');

	var MaxFragSize = 0xFFFFFFFF;
	var fragmentBuf = new ArrayBuffer(0);
	var isFragmented = false;

	//returns location of the data
	function readHeader(dv,fips){
		/* https://msdn.microsoft.com/en-us/library/cc240621.aspx */
		if(dv.byteLength < 3){
			return;
		}
		var o = {},dataStart;
		if(dv.getUint8(1) & 0x80){
			o.byteLength = dv.getUint16(1,false) & 0x7FFF;
			if(o.byteLength > dv.byteLength){
				return;
			}
			dataStart = 3;
		}else{
			o.byteLength = dv.getUint8(1);
			dataStart = 2;
		}


		var fpOutputHeader = dv.getUint8(0);
		if(fpOutputHeader & 0x03 !== 0x00){
			throw new Error('Invalid Action');
		}
		if(fpOutputHeader & 0x3C !== 0x00){
			throw new Error('Output header bits are reserved');
		}
		if(fpOutputHeader & 0x08){//secure checksum -- salted Mac
			o.salt = true;
		}
		if(fips){
			o.fips = new DataView(dv.buffer,dv.byteOffset + dataStart,4);
			dataStart +=4;
		}
		if(fpOutputHeader & 0x04){//encrypted
			o.signature = new DataView(dv.buffer,dv.byteOffset + dataStart,8);
			dataStart += 8;
		}
		o.data = new DataView(dv.buffer,dv.byteOffset + dataStart,o.byteLength - dataStart);
		return o;
	}
	//return bytes read
	function readUpdateHeader(dv){
		var o = {},size,sizeOffset = 1;
		o.updateCode = dv.getUint8(0) & 0x0F;
		o.fragmentation = (dv.getUint8(0) >>> 4) & 0x03;

		if(dv.getUint8(0) & 0xC0 === 0x02){
			o.compressionFlags = dv.getUint8(1);
			sizeOffset = 2;
		}

		size = dv.getUint16(sizeOffset,true);
		if(size){
			o.data = new DataView(dv.buffer,dv.byteOffset + sizeOffset + 2,size);
		}
		return o;
	}
	function processData(dv){
		var updateHeader = readUpdateHeader(dv);

		if(!updateHeader){
			throw new Error('Error reading update Header');
		}
		readUpdateData(updateHeader);
	}

	function readFragmentData(dv,fragCode){
		switch(fragCode){
			case 0x00://Fragment_single
				return dv;
				break;
			case 0x01://Fragment_last
				isFragmented = false;
				if(dv.byteLength + fragmentBuf.byteLength > MaxFragSize){
					throw new Error('Fragmentation size exceeded');
				}
				var tmpBuf = new ArrayBuffer(dv.byteLength + fragmentBuf.byteLength);
				(new Uint8Array(tmpBuf)).set(new Uint8Array(fragmentBuf),0);
				(new Uint8Array(tmpBuf)).set(new Uint8Array(dv.buffer,dv.byteOffset,dv.byteLength),fragmentBuf.byteLength);
				fragmentBuf = tmpBuf;
				return new DataView(fragmentBuf);
				break;
			case 0x02://fragment_first
				isFragmented = true;
				if(dv.byteLength > MaxFragSize){
					throw new Error('Fragmentation size exceeded');
				}
				fragmentBuf = new ArrayBuffer(dv.byteLength);
				(new Uint8Array(fragmentBuf)).set(new Uint8Array(dv.buffer,dv.byteOffset,dv.byteLength));
				return;
				break;
			case 0x03://fragment_next
				if(dv.byteLength + fragmentBuf.byteLength > MaxFragSize){
					throw new Error('Fragmentation size exceeded');
				}
				var tmpBuf = new ArrayBuffer(dv.byteLength + fragmentBuf.byteLength);
				(new Uint8Array(tmpBuf)).set(new Uint8Array(fragmentBuf),0);
				(new Uint8Array(tmpBuf)).set(new Uint8Array(dv.buffer,dv.byteOffset,dv.byteLength),fragmentBuf.byteLength);
				fragmentBuf = tmpBuf;
				return;
				break;
			default:
				throw new Error('Unknown Fragmentation');
		}
	}
	function setMaxFragSize(mfs){
		MaxFragSize = mfs;
	}


	/*************** Update Commands *****************/
    var frameNum =0;
	function readUpdateData(updateHeader){
		var dv;
		if(updateHeader.fragmentation === 0x00){
			//not fragmented
			dv = updateHeader.data;
		}else{
			dv = readFragmentData(updateHeader.data,updateHeader.fragmentation);
		}
		if(dv){
			switch(updateHeader.updateCode){
				case 0x00:
					var order;
					var orders =  Orders.process(dv);
					while(order = orders.shift()){
						CanvasOrder.process(order,frameNum++);
					}
					break;
				case 0x01://bitmap
					Bitmap.process(dv);
					break;
				case 0x02:
					Palette.process(dv);
					break
				case 0x03:
					//sync updates are obsolete and should be ignored
					break;
				case 0x04:
					Surface.readSufaceCommand(dv,frameNum++);
					break;
				case 0x05://system pointer hidden update
					PointerUpdate.pointerHidden(dv);
					break;
				case 0x06://system pointer default update
					PointerUpdate.pointerDefault(dv);
				case 0x08:
					PointerUpdate.pointerPosition(dv);
				case 0x09:
					PointerUpdate.colorPointer(dv);
					break;
				case 0x0A:
					PointerUpdate.cachedPointer(dv);
				case 0x0B:
					PointerUpdate.newPointer(dv);
					break;
				default:
					console.log('unknown update code',updateHeader);
					break;
			}
		}
	}

	function FP(Client,canvas){
		MaxFragSize = Client.store.serverCapabilities[26].maxRequestSize;
		return {
			readHeader:readHeader,
			process:processData,
		};
	}
module.exports = FP;
