'use strict';
/*
 * RDP5 bulk compression.
 */

var ForwardReadStream       = require("xrdp/rfx/decompress/js/ForwardReadStream");
var ForwardWriteStream      = require("xrdp/rfx/decompress/js/ForwardWriteStream");

var PACKET_COMPR_TYPE_64K = 0x1;

var PACKET_COMPRESSED = 0x20;
var PACKET_AT_FRONT = 0x40;
var PACKET_FLUSHED = 0x80;

function encodeLiteral(ws, byte) {
	if (byte > 0x7F) {
		ws.write(2,0x2); // 10
		ws.write(7,byte);
	}
	else ws.write(8,byte);
}

function encodeTuple(ws, co, lom) {
	if (co < 64) {
		ws.write(5,0x1F); // 11111
		ws.write(6,co);
	}
	else if (co < 320) {
		ws.write(5,0x1E); // 11110
		ws.write(8,co - 64);
	}
	else if (co < 2368) {
		ws.write(4,0xE); // 1110
		ws.write(11,co - 320);
	}
	else if (co < 65536) {
		ws.write(3,0x6); // 110
		ws.write(16,co - 2368);
	}

	if (lom == 3) {
		ws.write(1,0);
	}
	else if (lom < 8) {
		ws.write(2,0x2); // 10
		ws.write(2,lom);
	}
	else if (lom < 16) {
		ws.write(3,0x6); // 110
		ws.write(3,lom);
	}
	else if (lom < 32) {
		ws.write(4,0xE); // 1110
		ws.write(4,lom);
	}
	else if (lom < 64) {
		ws.write(5,0x1E); // 11110
		ws.write(5,lom);
	}
	else if (lom < 128) {
		ws.write(6,0x3E); // 111110
		ws.write(6,lom);
	}
	else if (lom < 256) {
		ws.write(7,0x7E); // 1111110
		ws.write(7,lom);
	}
	else if (lom < 512) {
		ws.write(8,0xFE); // 11111110
		ws.write(8,lom);
	}
	else if (lom < 1024) {
		ws.write(9,0x1FE); // 111111110
		ws.write(9,lom);
	}
	else if (lom < 2048) {
		ws.write(10,0x3FE); // 1111111110
		ws.write(10,lom);
	}
	else if (lom < 4096) {
		ws.write(11,0x7FE); // 11111111110
		ws.write(11,lom);
	}
	else if (lom < 8192) {
		ws.write(12,0xFFE); // 111111111110
		ws.write(12,lom);
	}
	else if (lom < 16384) {
		ws.write(13,0x1FFE); // 1111111111110
		ws.write(13,lom);
	}
	else if (lom < 32768) {
		ws.write(14,0x3FFE); // 11111111111110
		ws.write(14,lom);
	}
	else if (lom < 65536) {
		ws.write(15,0x7FFE); // 111111111111110
		ws.write(15,lom);
	}
}

var RDP5BulkDecompressor = function() {
	this.historyBuffer = new Uint8Array(65536);
	this.historyOffset = 0;
}

RDP5BulkDecompressor.prototype.decompress = function(flags, compressed, bytes) {
	if (!(flags & PACKET_COMPR_TYPE_64K)) return undefined;
	if (flags & PACKET_AT_FRONT && !(flags & PACKET_COMPRESSED)) return undefined;

	if (flags & PACKET_FLUSHED) {
		while (this.historyOffset--) this.historyBuffer[this.historyOffset] = 0;
		this.historyOffset = 0;
	}
	else if (flags & PACKET_AT_FRONT) {
		while (this.historyOffset--) this.historyBuffer[this.historyOffset] = 0;
		this.historyOffset = 0;
	}

	var ua = new Uint8Array(bytes);

	if (!(flags & PACKET_COMPRESSED)) {
		ua.set(new Uint8Array(compressed));
		return compressed.byteLength;
	}

	var rs = new ForwardReadStream(compressed);
	var index = 0;

	while (index < ua.length) {
		var value = 0;
		var isLit = false;

		switch (rs.readBoundedRun(1,5)) {
		case 0:
			value = rs.read(7);
			isLit = true;
			break;
		case 1:
			value = 0xF0 | rs.read(7);
			isLit = true;
			break;
		case 2:
			value = rs.read(16) + 2368;
			break;
		case 3:
			value = rs.read(11) + 320;
			break;
		case 4:
			value = rs.read(8) + 64;
			break;
		case 5:
			value = rs.read(6);
			break;
		}

		if (isLit) {
			if (this.historyOffset >= this.historyBuffer.length) return undefined;

			this.historyBuffer[this.historyOffset++] = ua[index++] = value;
		}
		else {
			if (!value) return undefined;
			if (value > this.historyOffset) return undefined;

			var run = rs.readBoundedRun(1,15);
			var lom = 0;

			if (run == 0) {
				lom = 3;
			}
			else if (run < 15) {
				lom = (1 << (run + 1)) | rs.read(run + 1);
			}

			if (lom < 3) return undefined;
			if (this.historyOffset + lom >= this.historyBuffer.length) return undefined;

			var cPtr = this.historyOffset - value;

			for (var i = 0; i < lom; i++) {
				this.historyBuffer[this.historyOffset++] = ua[index++] = this.historyBuffer[cPtr++];
			}
		}
	}

	return index;
}

var RDP5BulkCompressor = function() {
	this.historyBuffer = new Uint8Array(65536);
	this.historyOffset = 0;
	this.historyPtr = 0;
	this.prefixes = {};
	this.nextPrefix = 0;
}

RDP5BulkCompressor.prototype.historyUpdated = function() {
	while (this.nextPrefix + 2 < this.historyPtr) {
		var h = (this.historyBuffer[this.nextPrefix] << 16)
			| (this.historyBuffer[this.nextPrefix + 1] << 8)
			| this.historyBuffer[this.nextPrefix + 2];
		var matches = h in this.prefixes
			? this.prefixes[h]
			: this.prefixes[h] = [];

		matches.unshift(this.nextPrefix++);

		if (matches.length > 10) matches.pop();
	}
}

RDP5BulkCompressor.prototype.findMatch = function() {
	var bidx = this.historyPtr;
	var blen = 0;

	if (this.historyPtr + 2 < this.historyOffset) {
		var h = (this.historyBuffer[this.historyPtr] << 16)
			| (this.historyBuffer[this.historyPtr + 1] << 8)
			| this.historyBuffer[this.historyPtr + 2];

		if (h in this.prefixes) {
			for (var i = 0; i < this.prefixes[h].length; i++) {
				var idx = this.prefixes[h][i];
				var len = 3;

				while (this.historyPtr + len < this.historyOffset
					&& this.historyBuffer[idx + len] == this.historyBuffer[this.historyPtr + len]) len++;

				if (len > blen) {
					bidx = idx;
					blen = len;
				}
			}
		}
	}

	return [ bidx, blen ];
}

RDP5BulkCompressor.prototype.compress = function(bytes) {
	var flags = PACKET_COMPR_TYPE_64K;

	if (this.historyOffset + bytes.byteLength > this.historyBuffer.length) {
		while (this.historyOffset--) this.historyBuffer[this.historyOffset] = 0;
		this.historyPtr = this.historyOffset = 0;
		this.prefixes = {};
		this.nextPrefix = 0;

		flags |= PACKET_AT_FRONT;
	}
	else if (this.historyOffset == 0) {
		flags |= PACKET_AT_FRONT;
	}

	this.historyBuffer.set(new Uint8Array(bytes),this.historyOffset);
	this.historyOffset += bytes.byteLength;

	var ws = new ForwardWriteStream;

	while (this.historyPtr < this.historyOffset) {
		this.historyUpdated();

		var nm = this.findMatch();
		var co = this.historyPtr - nm[0];
		var lom = nm[1];

		if (co) {
			encodeTuple(ws,co,lom);
			this.historyPtr += lom;
		}
		else {
			encodeLiteral(ws,this.historyBuffer[this.historyPtr]);
			this.historyPtr++;
		}

		if (ws.byteLength >= bytes.byteLength) {
			while (this.historyOffset--) this.historyBuffer[this.historyOffset] = 0;
			this.historyPtr = this.historyOffset = 0;
			this.prefixes = {};
			this.nextPrefix = 0;

			flags &= ~PACKET_AT_FRONT;

			return [ flags | PACKET_FLUSHED, bytes ];
		}
	}

	return [ flags | PACKET_COMPRESSED, ws.buffer ];
}

module.exports = {
	decompressor: RDP5BulkDecompressor,
	compressor: RDP5BulkCompressor
}
